from typing import Any

NOT_NULL = ...  # type: Any
PRI_KEY = ...  # type: Any
UNIQUE_KEY = ...  # type: Any
MULTIPLE_KEY = ...  # type: Any
BLOB = ...  # type: Any
UNSIGNED = ...  # type: Any
ZEROFILL = ...  # type: Any
BINARY = ...  # type: Any
ENUM = ...  # type: Any
AUTO_INCREMENT = ...  # type: Any
TIMESTAMP = ...  # type: Any
SET = ...  # type: Any
PART_KEY = ...  # type: Any
GROUP = ...  # type: Any
UNIQUE = ...  # type: Any
